"""Docker network-info scan plugin."""
from DockerENT.utils import utils

import docker
import logging

_log = logging.getLogger(__name__)

_plugin_name_ = 'netinfo'


def scan(container, output_queue, audit=False, audit_queue=None):
    """Docker network-info plugin scan.

    :param container: container instance.
    :type container: docker.models.containers.Container

    :param output_queue: Output holder for this plugin.
    :type output_queue: multiprocessing.managers.AutoProxy[Queue]

    :return: This plugin returns the object in this form.
    {
        _plugin_name_: {
            'test_performed': {
                'results':  []
            }
        }
    }
    """
    res = {}

    _log.info('Staring {} Plugin ...'.format(_plugin_name_))

    api_client = docker.APIClient()
    docker_inspect_output = api_client.inspect_container(container.short_id)

    netinfo = {
        "NETINFO": {
            "isDockerInspectAttribute": False,
            "cmd": "/sbin/ifconfig -a",
            "msg": "Interfaces",
            "results": []
        },
        "ROUTE": {
            "isDockerInspectAttribute": False,
            "cmd": "route",
            "msg": "Route(s)",
            "results": []
        },
        "NETSTAT": {
            "isDockerInspectAttribute": False,
            "cmd": "netstat -antup",
            "msg": "Netstat",
            "results": []
        },
        "PORT_BINDINGS": {
            "isDockerInspectAttribute": True,
            "attribute_name": "NetworkSettings.Ports",
            "msg": "Docker port bindings",
            "results": []
        }
    }

    for item in netinfo.keys():
        if not netinfo[item]['isDockerInspectAttribute']:
            cmd = netinfo[item]['cmd']
            output = container.exec_run(cmd).output.decode('utf-8')
            result = output.split('\n')
            netinfo[item]['results'] = result
            del netinfo[item]['cmd']
        else:
            if netinfo[item]['isDockerInspectAttribute']:
                option_result = utils.get_value_from_str_dotted_key(
                    docker_inspect_output,
                    netinfo[item]['attribute_name']
                )
                del netinfo[item]['attribute_name']

                # Key not found
                if option_result is None:
                    del netinfo[item]['isDockerInspectAttribute']
                    continue
            # Output of docker inspect port is a map, with keys as
            netinfo[item]['results'].append(option_result)

        del netinfo[item]['isDockerInspectAttribute']

    result = netinfo

    res[container.short_id] = {
        _plugin_name_: result
    }

    _log.info('Completed execution of {} Plugin.'.format(_plugin_name_))
    output_queue.put(res)

    if audit:
        _audit(container, netinfo, audit_queue)


def _audit(container, scan_report, audit_queue):
    """Perform Scan audit.

    :param scan_report: dict
    :param audit_queue: Multiprocessing queue to perform Audit.
    """

    container_id = container.short_id
    audit_report = {}
    audit_report[container_id] = []

    columns = [_plugin_name_, 'INFO']

    weak_configurations = {
        'PORT_BINDINGS': {
            'safe_conf': [],
            'warn_msg': 'Port Mapping found:'
        }
    }

    for security_option in scan_report.keys():
        if security_option in weak_configurations.keys():
            actual_results = scan_report[security_option]['results']
            weak_results = None
            safe_results = None

            if 'weak_conf' in weak_configurations[security_option].keys():
                weak_results = weak_configurations[security_option]['weak_conf']

            if 'safe_conf' in weak_configurations[security_option].keys():
                safe_results = weak_configurations[security_option]['safe_conf']

            warn_message = weak_configurations[security_option]['warn_msg']

            # If safe_conf, check on safe_conf else check on weak_conf
            if safe_results is not None:
                if actual_results != safe_results:
                    _r = [res for res in actual_results if res]
                    if security_option == 'PORT_BINDINGS':
                        if _r:
                            r = columns + [warn_message + ";".join(
                                utils.docker_network_response_parser(_r)
                            )]
                            audit_report[container_id].append(", ".join(r))
            else:
                if utils.list_intersection(actual_results, weak_results) != []:
                    # Weak configuration detected
                    r = columns + [warn_message]
                    audit_report[container_id].append(
                        ", ".join(r)
                    )

    audit_queue.put(audit_report)



"""Docker file-system-info scan plugin."""
from DockerENT.utils import utils

import docker
import logging
import json

_log = logging.getLogger(__name__)

_plugin_name_ = 'files-info'


def scan(container, output_queue, audit=False, audit_queue=None):
    """Docker file-system-info plugin scan.

    :param container: container instance.
    :type container: docker.models.containers.Container

    :param output_queue: Output holder for this plugin.
    :type output_queue: multiprocessing.managers.AutoProxy[Queue]

    :return: This plugin returns the object in this form.
    {
        _plugin_name_: {
            'test_performed': {
                'results':  []
            }
        }
    }
    """
    res = {}

    _log.info('Starting {} Plugin ...'.format(_plugin_name_))

    api_client = docker.APIClient()
    docker_inspect_output = api_client.inspect_container(container.short_id)

    fileinfo = {
        "WWDIRSROOT": {
            "isDockerAttribute": False,
            "cmd": "find /  -wholename '/home/' -prune  -user root -o  -type d -perm -0002 -exec ls -ld '{}' ';'",
            "msg": "World Writeable Directories for User/Group 'Root'",
            "results": []
        },
        "WWDIRS": {
            "isDockerAttribute": False,
            "cmd": "find / \( -wholename '/home/' -prune \) -user root -o \( -type d -perm -0002 \) -exec ls -ld '{}' ';'",
            "msg": "World Writeable Directories for Users other than Root",
            "results": []
        },
        "WWFILES": {
            "isDockerAttribute": False,
            "cmd": "find / \( -wholename '/home/' -prune -user root -o -wholename '/proc/' -prune \) -o \( -type f -perm -0002 \) -exec ls -l '{}' ';'",
            "msg": "World Writable Files",
            "results": []
        },
        "SUID": {
            "isDockerAttribute": False,
            "cmd": "find / \( -perm -2000 -o -perm -4000 \) -exec ls -ld {} \;",
            "msg": "SUID/SGID Files and Directories",
            "results": []
        },
        "ROOTHOME": {
            "isDockerAttribute": False,
            "cmd": "ls -ahlR /root",
            "msg": "Checking if root's home folder is accessible",
            "results": []
        }
    }

    for item in fileinfo.keys():
        if fileinfo[item]['isDockerAttribute']:
            option_result = utils.get_value_from_str_dotted_key(
                docker_inspect_output,
                fileinfo[item]['attribute_name']
            )
            del fileinfo[item]['attribute_name']

            # Key not found
            if option_result is None:
                del fileinfo[item]['isDockerAttribute']
                continue
            elif isinstance(option_result, list):
                fileinfo[item]['results'].extend(option_result)
            else:
                fileinfo[item]['results'].append(option_result)

        else:
            cmd = fileinfo[item]['cmd']

            output = container.exec_run(cmd).output.decode('utf-8')

            result = output.split('\n')
            fileinfo[item]['results'] = result
            del fileinfo[item]['cmd']

        del fileinfo[item]['isDockerAttribute']

    result = fileinfo

    res[container.short_id] = {
        _plugin_name_: result
    }

    _log.info('Completed execution of {} Plugin.'.format(_plugin_name_))
    output_queue.put(res)

    if audit:
        # Perform Audit for this result
        _audit(container, fileinfo, audit_queue)


def _audit(container, scan_report, audit_queue):
    """Perform Scan audit.

    :param scan_report: dict
    :param audit_queue: Multiprocessing queue to perform Audit.
    """
    weak_configurations = {
        "WWDIRSROOT": {
            "safe_conf": [],
            "warn_msg": "World Writeable Directories for User/Group 'Root': ",
        },
        "WWDIRS": {
            "safe_conf": [],
            "warn_msg": "World Writeable Directories for Users other than Root",
        },
        "WWFILES": {
            "safe_conf": [],
            "warn_msg": "World Writable Files: ",
        },
        "SUID": {
            "safe_conf": [],
            "warn_msg": "SUID/SGID Files and Directories: ",
        },
        "ROOTHOME": {
            "safe_conf": [],
            "warn_msg": "Checking if root's home folder is accessible: ",
        }
    }

    container_id = container.short_id
    audit_report = {}
    audit_report[container_id] = []

    columns = [_plugin_name_, 'WARN']
    for security_option in scan_report.keys():
        if security_option in weak_configurations.keys():
            actual_results = scan_report[security_option]['results']
            if isinstance(actual_results, dict):
                actual_results = json.dumps(actual_results)

            weak_results = None
            safe_results = None

            if 'weak_conf' in weak_configurations[security_option].keys():
                weak_results = weak_configurations[security_option]['weak_conf']

            if 'safe_conf' in weak_configurations[security_option].keys():
                safe_results = weak_configurations[security_option]['safe_conf']

            warn_message = weak_configurations[security_option]['warn_msg']

            # If safe_conf, check on safe_conf else check on weak_conf
            if safe_results is not None:
                if actual_results != safe_results:
                    _r = [res for res in actual_results if res]
                    if _r:
                        r = columns + [warn_message + ";".join(_r)]
                        audit_report[container_id].append(", ".join(r))
            else:
                for wr in weak_results:
                    if any(wr in ar for ar in actual_results):
                        # Weak configuration detected
                        r = columns + [warn_message]
                        audit_report[container_id].append(
                            ", ".join(r)
                        )

    audit_queue.put(audit_report)



"""Docker security profiles scan plugin."""
from DockerENT.utils import utils

import docker
import logging

_log = logging.getLogger(__name__)

_plugin_name_ = 'security-profiles'


def scan(container, output_queue, audit=False, audit_queue=None):
    """Docker security profiles scan plugin.

    This plugin inspect the AppArmor and SELinux profile for running
    container via low level docker python sdk low level api.

    :param container: container instance.
    :type container: docker.models.containers.Container

    :param output_queue: Output holder for this plugin.
    :type output_queue: multiprocessing.managers.AutoProxy[Queue]

    :return: This plugin returns the object in this form.
    {
        _plugin_name_: {
            'test_performed': {
                'results':  []
            }
        }
    }
    """
    res = {}

    _log.info('Staring {} Plugin ...'.format(_plugin_name_))

    api_client = docker.APIClient()
    docker_inspect_output = api_client.inspect_container(container.short_id)

    security_options = {
        "AppArmor_Profile": {
            "output_identifier": "AppArmorProfile",
            "msg": "AppArmor Profile options.",
            "results": []
        },
        "SELinux": {
            "output_identifier": "HostConfig.SecurityOpt",
            "msg": "SELinux security options.",
            "results": []
        },
        "CapAdd": {
            "output_identifier": "HostConfig.CapAdd",
            "msg": "A list of kernel capabilities to added to the container.",
            "results": []
        },
        "CapDrop": {
            "output_identifier": "HostConfig.CapDrop",
            "msg": "A list of kernel capabilities to drop from the container.",
            "results": []
        },
        "Capabilities": {
            "output_identifier": "HostConfig.Capabilities",
            "msg": "A list of kernel capabilities present container.",
            "results": []
        },
        "Privileged": {
            "output_identifier": "HostConfig.Privileged",
            "msg": " Gives the container full access to the host.",
            "results": []
        },

    }

    for option in security_options.keys():
        option_result = utils.get_value_from_str_dotted_key(
            docker_inspect_output,
            security_options[option]['output_identifier']
        )
        if isinstance(option_result, list):
            security_options[option]['results'] = option_result
        elif option_result is None:
            security_options[option]['results'] = [None]
        else:
            security_options[option]['results'].append(option_result)

        del security_options[option]['output_identifier']

    res[container.short_id] = {
        _plugin_name_: security_options
    }

    _log.info('Completed execution of {} Plugin.'.format(_plugin_name_))
    output_queue.put(res)

    if audit:
        # Perform Audit for this result
        _audit(container, security_options, audit_queue)


def _audit(container, scan_report, audit_queue):
    """Perform Scan audit.

    :param scan_report: dict
    :param audit_queue: Multiprocessing queue to perform Audit.
    """
    container_id = container.short_id
    audit_report = {}
    audit_report[container_id] = []

    columns = [_plugin_name_, 'WARN']

    weak_configurations = {
        'AppArmor_Profile': {
            'weak_conf': ['', None, '<no value>', 'unconfined'],
            'warn_msg': 'No AppArmorProfile Found'
        },
        'SELinux': {
            'weak_conf': ['', None, '<no value>', 'unconfined',
                          'label=disable'],
            'warn_msg': 'SELinux disabled'
        },
        'CapAdd': {
            'safe_conf': [],
            'warn_msg': 'Capabilities Present: '
        },
        'Privileged': {
            'weak_conf': ['true', 'TRUE', True],
            'warn_msg': 'Container is Privileged'
        }
    }

    for security_option in scan_report.keys():
        if security_option in weak_configurations.keys():
            actual_results = scan_report[security_option]['results']
            weak_results = None
            safe_results = None

            if 'weak_conf' in weak_configurations[security_option].keys():
                weak_results = weak_configurations[security_option]['weak_conf']

            if 'safe_conf' in weak_configurations[security_option].keys():
                safe_results = weak_configurations[security_option]['safe_conf']

            warn_message = weak_configurations[security_option]['warn_msg']

            # If safe_conf, check on safe_conf else check on weak_conf
            if safe_results is not None:
                if actual_results != safe_results:
                    _r = [res for res in actual_results if res]
                    if _r:
                        r = columns + [warn_message + ";".join(_r)]
                        audit_report[container_id].append(", ".join(r))
            else:
                if utils.list_intersection(actual_results, weak_results) != []:
                    # Weak configuration detected
                    r = columns + [warn_message]
                    audit_report[container_id].append(
                        ", ".join(r)
                    )

    audit_queue.put(audit_report)



"""The package for output plugins packaged with this project.

This package contains output plugins which are packaged with this project.
The output plugins implements a function ``write`` that accepts the queue of
type multiprocessing.managers.AutoProxy[Queue] and process the elements from
the queue and create output pipeline.
"""



"""Config & Config parser for DockerENT."""
import yaml

config_yml = """
logger:
  version: 1
  disable_existing_loggers: false
  formatters:
    simple:
      format: >-
          %(asctime)s [%(process)s] [%(processName)s] [%(threadName)s]
          %(levelname)s %(name)s:%(lineno)d - %(message)s
      datefmt: "%Y-%m-%d %H:%M:%S"
  handlers:
    rich_console:
      class: rich.logging.RichHandler
      formatter: simple
    console:
      class: logging.StreamHandler
      formatter: simple
      stream: ext://sys.stdout
  loggers:
    adal-python:
      level: WARNING
  root:
    level: INFO
    handlers:
      - rich_console
"""
config = yaml.safe_load(config_yml)



"""The controller module, the main method of this module starts DockerENT."""
from DockerENT import output_worker
from DockerENT import scanner_workers
from multiprocessing import pool

import logging
import multiprocessing

# Define module-level logger.
_log = logging.getLogger(__name__)


def main(docker_containers,
         docker_plugins,
         docker_nws,
         docker_nw_plugins,
         process_count,
         audit,
         output):
    """Start DockerENT application.

    :return: None
    """
    _log.info('Starting application ...')

    _log.info('Creating application pool space with count {}'
              .format(process_count))

    process_pool = pool.Pool(process_count)
    output_q = multiprocessing.Manager().Queue()
    audit_q = None

    if audit:
        audit_q = multiprocessing.Manager().Queue()

    if docker_containers is not None:
        scanner_workers.docker_scan_worker(
            containers=docker_containers,
            plugins=docker_plugins,
            process_pool=process_pool,
            output_queue=output_q,
            audit=audit,
            audit_queue=audit_q
        )

    if docker_nws is not None:
        scanner_workers.docker_nw_scan_worker(
            nws=docker_nws,
            plugins=docker_nw_plugins,
            process_pool=process_pool,
            output_queue=output_q
        )

    process_pool.close()
    process_pool.join()

    output_worker.output_handler(
        queue=output_q,
        target=output)

    if audit:
        output_worker.output_handler(
            queue=audit_q,
            target=output,
            filename='audit.json'
        )



"""HTML output plugin.

Process the records and write the content to the html file which can show
docker connects as networks. This file uses a templates called as template.html
"""
import jinja2
import json
import logging

_log = logging.getLogger(__name__)

_plugin_name_ = 'HTML_output_plugin'


def write(queue, filename='out.html'):
    """Plugin to create beautiful html file for output.

    :param queue: The input queue which contains the data from various scan
        plugins.
    :type queue: multiprocessing.managers.AutoProxy[Queue]

    :param filename: Name of file to output to.
    :type filename: str

    :return: None
    """
    _log.info('Writing to {} ...'.format(filename))

    nodes = []
    edges = []

    report = {}
    while not queue.empty():
        result = queue.get()
        for key in result.keys():
            if key in report.keys():
                report[key].append(result[key])
            else:
                report[key] = []
                report[key].append(result[key])

    """
    After processing the list would look something like this.
    {
      "3911e81b25": [
        {
          "sample": {
            "status": "good"
          }
        }
      ]
    }
    """
    group = -1
    for docker_node in report.keys():
        group += 1
        nodes.append(
            {
                'id': docker_node,
                'shape': 'image',
                'image': 'https://www.docker.com/sites/default/files/d8/2019'
                         '-07/Moby-logo.png',
                'label': docker_node,
                'pluginname': docker_node,
                'data': 'This is docker node',
                'labelHighlightBold': True,
            }
        )
        for plugin_node in report[docker_node]:
            for plugin_data_node in plugin_node.keys():
                node_id = 'plugin_' + plugin_data_node + '_' + docker_node
                nodes.append(
                    {
                        'id': node_id,
                        'shape': 'dot',
                        'label': plugin_data_node,
                        'pluginname': plugin_data_node,
                        'data': 'This is plugin node',
                        'labelHighlightBold': True,

                    }
                )

                edges.append(
                    {
                        'from': docker_node,
                        'to': node_id,
                        'length': 200
                    }
                )

                for cmd_node in plugin_node[plugin_data_node].keys():
                    child_node_id = docker_node + '_' + cmd_node
                    nodes.append(
                        {
                            'id': child_node_id,
                            'shape': 'dot',
                            'label': cmd_node,
                            'pluginname': cmd_node,
                            'data': plugin_node[plugin_data_node][cmd_node][
                                'results'],
                            'labelHighlightBold': True,
                            'group': group,
                        }
                    )
                    edges.append(
                        {
                            'from': node_id,
                            'to': child_node_id,
                            'value': 5,
                            'length': 100
                        }
                    )

    with open(filename, 'w+') as outfile, \
            open('DockerENT/output_plugins/template.html', 'r') as template:

        template_html = jinja2.Template(template.read())
        html = template_html.render(
            nodes=json.dumps(nodes),
            edges=json.dumps(edges)
        )
        outfile.write(html)



"""Docker system info scan plugin."""
import logging

_log = logging.getLogger(__name__)

_plugin_name_ = 'sysinfo'


def scan(container, output_queue, audit=False, audit_queue=None):
    """Docker system-info scan plugin.

    :param container: container instance.
    :type container: docker.models.containers.Container

    :param output_queue: Output holder for this plugin.
    :type output_queue: multiprocessing.managers.AutoProxy[Queue]

    :return: This plugin returns the object in this form.
    {
        _plugin_name_: {
            'test_performed': {
                'results':  []
            }
        }
    }
    """
    res = {}

    _log.info('Staring {} Plugin ...'.format(_plugin_name_))

    sysinfo = {
        "OS": {
            "cmd": "cat /etc/issue",
            "msg": "Operating System",
            "results": []
        },
        "KERNEL": {
            "cmd": "cat /proc/version",
            "msg": "Kernel",
            "results": []
        },
        "HOSTNAME": {
            "cmd": "hostname",
            "msg": "Hostname",
            "results": []}
    }

    for item in sysinfo:
        cmd = sysinfo[item]['cmd']
        output = container.exec_run(cmd).output.decode('utf-8')

        sysinfo[item]['results'] = output
        del sysinfo[item]['cmd']

    result = sysinfo

    res[container.short_id] = {
        _plugin_name_: result
    }

    _log.info('Completed execution of {} Plugin.'.format(_plugin_name_))
    output_queue.put(res)



"""File output plugin.

Process the records and write the content to the file as plain json.
"""
import json
import logging

_log = logging.getLogger(__name__)

_plugin_name_ = 'file-output-plugin'


def write(queue, filename='out.json'):
    """Write report to to file.

    :param queue: The input queue which contains the data from various scan
        plugins.
    :type queue: multiprocessing.managers.AutoProxy[Queue]

    :param filename: Name of file to output to.
    :type filename: str

    :return: None
    """
    _log.info("Writing to file {} ...".format(filename))

    # Perform data massaging
    report = {}
    while not queue.empty():
        result = queue.get()
        for key in result.keys():
            if key in report.keys():
                report[key].append(result[key])
            else:
                report[key] = []
                report[key].append(result[key])

    if report:
        with open(filename, 'w+') as file:
            file.write(json.dumps(report, indent=2))

    _log.info("Written to disk {}".format(filename))



"""Docker file-system-info scan plugin."""
from DockerENT.utils import utils

import docker
import logging
import json

_log = logging.getLogger(__name__)

_plugin_name_ = 'file-system-info'


def scan(container, output_queue, audit=False, audit_queue=None):
    """Docker file-system-info plugin scan.

    :param container: container instance.
    :type container: docker.models.containers.Container

    :param output_queue: Output holder for this plugin.
    :type output_queue: multiprocessing.managers.AutoProxy[Queue]

    :return: This plugin returns the object in this form.
    {
        _plugin_name_: {
            'test_performed': {
                'results':  []
            }
        }
    }
    """
    res = {}

    _log.info('Starting {} Plugin ...'.format(_plugin_name_))

    api_client = docker.APIClient()
    docker_inspect_output = api_client.inspect_container(container.short_id)

    driveinfo = {
        "MOUNT": {
            "isDockerAttribute": False,
            "cmd": "mount",
            "msg": "Mount results",
            "results": []
        },
        "FSTAB": {
            "isDockerAttribute": False,
            "cmd": "cat /etc/fstab",
            "msg": "fstab entries",
            "results": []
        },
        "RW_VOLUMES": {
            "isDockerAttribute": True,
            "attribute_name": "VolumesRW",
            "msg": "R/W volumes mounted in docker.",
            "results": []

        },
        "DOCKER_MOUNTS": {
            "isDockerAttribute": True,
            "attribute_name": "Mounts",
            "msg": "List mounted volumes in docker.",
            "results": []
        }
    }

    for item in driveinfo.keys():
        if driveinfo[item]['isDockerAttribute']:
            option_result = utils.get_value_from_str_dotted_key(
                docker_inspect_output,
                driveinfo[item]['attribute_name']
            )
            del driveinfo[item]['attribute_name']

            # Key not found
            if option_result is None:
                del driveinfo[item]['isDockerAttribute']
                continue
            elif isinstance(option_result, list):
                driveinfo[item]['results'].extend(option_result)
            else:
                driveinfo[item]['results'].append(option_result)

        else:
            cmd = driveinfo[item]['cmd']
            output = container.exec_run(cmd).output.decode('utf-8')
            result = output.split('\n')
            driveinfo[item]['results'] = result
            del driveinfo[item]['cmd']

        del driveinfo[item]['isDockerAttribute']

    result = driveinfo

    res[container.short_id] = {
        _plugin_name_: result
    }

    _log.info('Completed execution of {} Plugin.'.format(_plugin_name_))
    output_queue.put(res)

    if audit:
        # Perform Audit for this result
        _audit(container, driveinfo, audit_queue)


def _audit(container, scan_report, audit_queue):
    """Perform Scan audit.

    :param scan_report: dict
    :param audit_queue: Multiprocessing queue to perform Audit.
    """
    weak_configurations = {
        "MOUNT": {
            'weak_conf': ['rw'],
            'warn_msg': 'READ-WRITE Mount'
        },
        "FSTAB": {
            'weak_conf': ['rw'],
            'warn_msg': 'READ-WRITE mount in fstab'
        },
        "RW_VOLUMES": {
            'weak_conf': ['rw'],
            'warn_msg': 'READ-WRITE Volume'
        },
        "DOCKER_MOUNTS": {
            'weak_conf': ['RW'],
            'warn_msg': 'RW in Docker Mount'
        }
    }
    container_id = container.short_id
    audit_report = {}
    audit_report[container_id] = []

    columns = [_plugin_name_, 'WARN']
    for security_option in scan_report.keys():
        if security_option in weak_configurations.keys():
            actual_results = scan_report[security_option]['results']
            if isinstance(actual_results, dict):
                actual_results = json.dumps(actual_results)

            weak_results = None
            safe_results = None

            if 'weak_conf' in weak_configurations[security_option].keys():
                weak_results = weak_configurations[security_option]['weak_conf']

            if 'safe_conf' in weak_configurations[security_option].keys():
                safe_results = weak_configurations[security_option]['safe_conf']

            warn_message = weak_configurations[security_option]['warn_msg']

            # If safe_conf, check on safe_conf else check on weak_conf
            if safe_results is not None:
                if actual_results != safe_results:
                    _r = [res for res in actual_results if res]
                    if _r:
                        r = columns + [warn_message + ";".join(_r)]
                        audit_report[container_id].append(", ".join(r))
            else:
                for wr in weak_results:
                    if any(wr in ar for ar in actual_results):
                        # Weak configuration detected
                        r = columns + [warn_message]
                        audit_report[container_id].append(
                            ", ".join(r)
                        )

    audit_queue.put(audit_report)



"""Web App for DockerENT."""
from DockerENT import scanner_workers
from DockerENT.utils import utils

import base64
import docker
import DockerENT
import json
import multiprocessing
import pkgutil
import streamlit as ui

ui_sidebar = ui.sidebar

# Title for Web application. Shown main web app
ui.title("DockerENT")

# Get the docker client to interact with docker api.
docker_client = docker.from_env()

# Get the list of all docker containers for sidebar drop down
docker_list = docker_client.containers.list()

# Convert container objects to string, for use by executor.
docker_list = [docker_name.short_id for docker_name in docker_list]
# Add option for 'all' to be shown in UI
docker_list.insert(0, 'all')

# Get the list of all docker-nw for sidebar drop down

docker_nw_list = docker_client.networks.list()
docker_nw_list = [docker_nw_name.short_id for docker_nw_name in docker_nw_list]
docker_nw_list.sort()
docker_nw_list.insert(0, 'all')

# --------------------------- SIDEBAR START -----------------------------------
# Initialise all UI Sidebar widgets
# Order is preserved in UI
ui_sidebar.text('Select options to start DockerENT')
scan_dockers_checkbox = ui_sidebar.checkbox('Scan dockers')
scan_docker_nw_checkbox = ui_sidebar.checkbox('Scan docker networks')

# Placeholders
docker_list_dropdown = None
docker_plugins_list_dropdown = None
docker_nw_list_dropdown = None
docker_nw_plugin_list_dropdown = None
ui_sidebar_start_dockerent = None
# --------------------------- SIDEBAR END -------------------------------------


# --------------------------- UI START ----------------------------------------
# Initialise all UI widgets
# Order is preserved in UI
ui_progress_bar = None
# --------------------------- UI END ------------------------------------------


# Global UI result variables
docker_scan_list = None
docker_scan_plugins = None
docker_scan_audit = None

docker_nw_scan_list = None
docker_nw_scan_plugins = None
docker_nw_scan_audit = None


class AutoUpdateProgressBar:
    """Progress bar autoupdate class."""

    def __init__(self, iterable, progress_bar):
        """Class method."""
        self.prog_bar = progress_bar
        self.iterable = iterable
        self.length = len(iterable)
        self.i = 0

    def __iter__(self):
        """Run on each iteration."""
        for obj in self.iterable:
            yield obj
            self.i += 1
            current_prog = self.i / self.length
            self.prog_bar.progress(current_prog)


def render_sidebar():
    """Render UI Sidebar."""
    global docker_scan_list
    global docker_scan_plugins
    global docker_scan_audit

    global docker_nw_scan_list
    global docker_nw_scan_plugins
    global docker_nw_scan_audit

    global ui_sidebar_start_dockerent

    # Docker UI Sidebar
    if scan_dockers_checkbox:
        docker_scan_list = ui_sidebar.multiselect(
            'Pick a single docker or all to start.',
            docker_list
        )

        _plugins = ['all']
        for importer, modname, ispkg in pkgutil.iter_modules(
                DockerENT.docker_plugins.__path__):
            _plugins.append(modname)

        docker_scan_plugins = ui_sidebar.multiselect(
            'Select the list of plugins to execute',
            _plugins
        )

        docker_scan_audit = ui_sidebar.checkbox('Audit Docker Results')

    # Docker Network UI Sidebar
    if scan_docker_nw_checkbox:
        docker_nw_scan_list = ui_sidebar.multiselect(
            'Pick Docker network to scan.',
            docker_nw_list
        )

        _plugins = ['all']
        for importer, modname, ispkg in pkgutil.iter_modules(
                DockerENT.docker_nw_plugins.__path__):
            _plugins.append(modname)

        docker_nw_scan_plugins = ui_sidebar.multiselect(
            'Select the list of plugins to execute',
            _plugins
        )

        docker_scan_audit = ui_sidebar.checkbox('Audit Docker N/W Results')

        if ui_sidebar_start_dockerent:
            scan_docker_networks()

    if scan_docker_nw_checkbox or scan_dockers_checkbox:
        ui_sidebar_start_dockerent = ui_sidebar.button('Start')

    if ui_sidebar_start_dockerent:
        if scan_dockers_checkbox:
            if docker_scan_list:
                scan_dockers()
            else:
                raise Exception('Please select one or more docker instance.')
        if scan_docker_nw_checkbox:
            if scan_docker_nw_checkbox:
                if docker_nw_scan_list:
                    scan_docker_networks()
                else:
                    raise Exception(
                        'Please select one or more Docker Networks')


def render_ui():
    """Render UI panel."""
    pass


@ui.cache
class Executor:
    def docker_scan_executor(self,
                             target,
                             plugin,
                             output_queue,
                             audit_queue):
        scanner_workers.executor(
            target=target,
            plugin=plugin,
            output_queue=output_queue,
            is_docker=True,
            audit=True,
            audit_queue=audit_queue
        )

    def docker_nw_scan_executor(self,
                                target,
                                plugin,
                                output_queue,
                                audit_queue):
        scanner_workers.executor(
            target=target,
            plugin=plugin,
            output_queue=output_queue,
            is_docker=False,
            audit_queue=audit_queue
        )


executor = Executor()


def scan_dockers():
    """Run DockerENT application on Dockers."""
    global docker_scan_list
    global docker_scan_plugins

    ui.markdown(
        """
        ### Docker Scan Section.
        """
    )
    # Create a Q to handle report from each plugin
    output_q = multiprocessing.Manager().Queue()
    audit_q = multiprocessing.Manager().Queue()

    _containers = []
    if docker_scan_list == 'all' or 'all' in docker_scan_list:
        _containers = docker_client.containers.list()
    else:
        for c in docker_scan_list:
            _containers.append(docker_client.containers.get(c))

    _plugins = []

    if not docker_scan_plugins:
        docker_scan_plugins = ['all']

    if docker_scan_plugins is None or 'all' in docker_scan_plugins \
            or docker_scan_plugins[0] == 'all':
        for importer, modname, ispkg in pkgutil.iter_modules(
                DockerENT.docker_plugins.__path__):
            _plugins.append(modname)
    else:
        _plugins = docker_scan_plugins

    docker_scan_progress_bar = ui.progress(0)
    # List of Tuples(target, plugin)to hold target to plugin combinations
    target_plugin = []

    for container in _containers:
        for plugin in _plugins:
            target_plugin.append((container, plugin))

    with ui.spinner('**Scanning** dockers ..'):
        for i in AutoUpdateProgressBar(range(len(target_plugin)),
                                       docker_scan_progress_bar):
            executor.docker_scan_executor(
                target=target_plugin[i][0].short_id,
                plugin=target_plugin[i][1],
                output_queue=output_q,
                audit_queue=audit_q
            )
    ui.success('**Docker Scan** Complete')

    ui.info('**Docker Scan Summary**')
    report = utils.parse_output_queue_2_dict(output_q)

    b64report = base64.b64encode(json.dumps(report).encode())
    href = f"""
    <a href="data:text/json;base64,{b64report.decode("utf-8")}" 
    download="report.json">Download RAW JSON report</a>"""

    ui.markdown(
        href,
        unsafe_allow_html=True
    )
    ui.json(report)

    if docker_scan_audit:
        ui.info('Docker Scan Audit Summary')

        audit_report = utils.parse_audit_queue_2_dict(audit_q)

        b64report = base64.b64encode(json.dumps(audit_report).encode())
        href = f"""
        <a href="data:text/json;base64,{b64report.decode("utf-8")}" 
        download="audit_report.json">Download Audit JSON report</a>"""

        ui.markdown(
            href,
            unsafe_allow_html=True
        )
        ui.json(audit_report)


def scan_docker_networks():
    """Run DockerENT application on Dockers NWs."""
    global docker_nw_scan_list
    global docker_nw_scan_plugins

    ui.markdown(
        """
        ### Docker Network Scan Section.
        """
    )

    # Create a Q to handle report from each plugin
    output_q = multiprocessing.Manager().Queue()
    audit_q = multiprocessing.Manager().Queue()

    _containers = []
    if docker_nw_scan_list == 'all' or 'all' in docker_nw_scan_list:
        _containers = docker_client.networks.list()
    else:
        for c in docker_nw_scan_list:
            _containers.append(docker_client.networks.get(c))

    _plugins = []

    if not docker_nw_scan_plugins:
        docker_nw_scan_plugins = ['all']

    if docker_nw_scan_plugins is None or 'all' in docker_nw_scan_plugins \
            or docker_nw_scan_plugins[0] == 'all':
        for importer, modname, ispkg in pkgutil.iter_modules(
                DockerENT.docker_nw_plugins.__path__):
            _plugins.append(modname)
    else:
        _plugins = docker_nw_scan_plugins

    docker_nw_scan_progress_bar = ui.progress(0)
    # List of Tuples(target, plugin)to hold target to plugin combinations
    target_plugin = []

    for container in _containers:
        for plugin in _plugins:
            target_plugin.append((container, plugin))

    with ui.spinner('Scanning docker networks ..'):
        for i in AutoUpdateProgressBar(range(len(target_plugin)),
                                       docker_nw_scan_progress_bar):
            executor.docker_nw_scan_executor(
                target=target_plugin[i][0].short_id,
                plugin=target_plugin[i][1],
                output_queue=output_q,
                audit_queue=audit_q
            )
    ui.success('Docker Networks Scan Complete')

    report = utils.parse_output_queue_2_dict(output_q)

    b64report = base64.b64encode(json.dumps(report).encode())
    href = f"""
    <a href="data:text/json;base64,{b64report.decode("utf-8")}" 
    download="report.json">Download RAW JSON report</a>"""

    ui.markdown(
        href,
        unsafe_allow_html=True
    )
    ui.json(report)

    if docker_nw_scan_audit:
        audit_report = {}
        ui.json(audit_report)


def main():
    """Start the UI Application."""
    render_sidebar()
    render_ui()


try:
    main()
except Exception as e:
    ui.error(str(e))



"""Main Script for DockerENT.

Script will load any config file and make it available to application.
"""
from DockerENT import config_parser
from DockerENT import controller
import DockerENT

import argparse
import logging.config
import os
import signal
import subprocess
import sys


def start():
    # Setup config for this package
    logger_config = config_parser.config['logger']
    logging.config.dictConfig(logger_config)

    _log = logging.getLogger(__name__)

    # Read arguments for DockerENT application
    parser = argparse.ArgumentParser(
        prog='Find the vulnerabilities hidden in your running container(s).'
    )

    parser.add_argument(
        '-v',
        '--version',
        dest='version',
        action='store_true',
        default=False,
        help='DockerENT version.'
    )

    docker_args_group = parser.add_argument_group()
    docker_args_group.add_argument(
        '-d',
        '--docker',
        nargs='?',
        dest='docker_container',
        const='all',
        help='Run scan against the running container.')
    docker_args_group.add_argument(
        '-p',
        '--plugins',
        nargs='?',
        dest='docker_plugins',
        const='all',
        help='Run scan with only specified plugins.')

    docker_nw_args_group = parser.add_argument_group()
    docker_nw_args_group.add_argument(
        '-d-nw',
        '--docker-network',
        nargs='?',
        dest='docker_network',
        const='all',
        help='Run scan against running docker-network.')
    docker_args_group.add_argument(
        '-p-nw',
        '--nw-plugins',
        nargs='?',
        dest='docker_nw_plugins',
        const='all',
        help='Run scan with only specified plugins.')

    parser.add_argument(
        '-w',
        '--web-app',
        dest='web_app',
        action='store_true',
        default=False,
        help='Run DockerENT in WebApp mode. '
        'If this parameter is enabled, other command line flags will be ignored.'
    )

    parser.add_argument(
        '-n',
        '--process',
        nargs='?',
        dest='process_count',
        default=2,
        type=int,
        help='Run scans in parallel (Process pool count).'
    )

    parser.add_argument(
        '-a',
        '--audit',
        dest='audit',
        action='store_true',
        default=False,
        help='Flag to check weather to audit results or not.'
    )

    output_plugin = parser.add_argument_group()
    output_plugin.add_argument(
        '-o',
        '--output',
        nargs='?',
        dest='output',
        default='file',
        type=str,
        help='Output plugin to write data to.'
    )

    args = parser.parse_args()

    if args.version:
        print(DockerENT.__version__)
        return

    process_count = args.process_count
    output = args.output
    audit = args.audit
    webapp = args.web_app

    docker_containers = args.docker_container
    docker_plugins = args.docker_plugins

    docker_nws = args.docker_network
    docker_nw_plugins = args.docker_nw_plugins

    # Register Signal Handler for graceful exit in case of Web application

    def sigterm_handler(_signo, _stack_frame):
        """Signal handler."""
        _log.info("Thanks for using DockerENT")
        sys.exit(0)

    signal.signal(signal.SIGINT, sigterm_handler)

    # Start DockerENT
    # If webapp, start Streamlit else cli application.
    if webapp:
        _log.info('Starting web application ...')

        web_app_file_name = os.path.abspath(DockerENT.__file__) + '/web_app.py'
        web_app_cmd = "streamlit run " + web_app_file_name

        with subprocess.Popen(web_app_cmd.split(" ")) as web_process:
            _log.info(web_process.stdout.read())

    else:
        controller.main(docker_containers=docker_containers,
                        docker_plugins=docker_plugins,
                        docker_nws=docker_nws,
                        docker_nw_plugins=docker_nw_plugins,
                        process_count=process_count,
                        audit=audit,
                        output=output)


start()



"""Docker network sample scan plugin."""
import logging

_log = logging.getLogger(__name__)

_plugin_name_ = 'sample_nw'


def scan(container, output_queue, audit=False, audit_queue=None):
    """Docker sample_nw scan plugin.

    :param container: container instance.
    :type container: docker.models.containers.Container

    :param output_queue: Output holder for this plugin.
    :type output_queue: multiprocessing.managers.AutoProxy[Queue]

    :return: This plugin returns the object in this form.
    {
        _plugin_name_: {
            'test_performed': {
                'results':  []
            }
        }
    }
    """
    _log.info('Staring {} Plugin ...'.format(_plugin_name_))

    res = {

    }
    result = {
        'test_performed': {
            'results': ['good']
        }
    }

    # do some processing.
    # Since this is sample plugin there is no processing.

    res[container.short_id] = {
        _plugin_name_: result
    }
    _log.info('Completed execution of {} Plugin.'.format(_plugin_name_))

    output_queue.put(res)



"""The package for docker scanner plugins packaged with this project.

This package contains docker-scanner plugins which are packaged with this
project. The docker-scanner plugins implements a function ``scan`` that
accepts a container instance of type ``docker.models.containers.Container``
and a queue of type ``multiprocessing.managers.AutoProxy[Queue]`` and then
executes the ``scan`` on that specific container.

``scan`` function should return a dict of below form for output plugins to work
    {
        _plugin_name_: {
            'test_performed': {
                'results':  []
            }
        }
    }
"""



"""Docker command history scan plugin."""
from DockerENT import utils

import logging

_log = logging.getLogger(__name__)

_plugin_name_ = 'user-command-history'


def scan(container, output_queue, audit=False, audit_queue=None):
    """Docker command history plugin scan.

    :param container: container instance.
    :type container: docker.models.containers.Container

    :param output_queue: Output holder for this plugin.
    :type output_queue: multiprocessing.managers.AutoProxy[Queue]

    :return: This plugin returns the object in this form.
    {
        _plugin_name_: {
            'test_performed': {
                'results':  []
            }
        }
    }
    """
    res = {}

    _log.info('Staring {} Plugin ...'.format(_plugin_name_))

    historyfiles = {
        "ROOT_SH_HISTORY": {
            "cmd": 'cat /root/.ash_history',
            "msg": "See if you have access too Root user history (depends on "
                   "privs)",
            "results": []
        },
        "ROOT_BASH_HISTORY": {
            "cmd": 'cat /root/.bash_history',
            "msg": "See if you have access too Root user history (depends on "
                   "privs)",
            "results": []
        },
        "USER_SH_HISTORY": {
            "cmd": 'cat ~/.ash_history ',
            "msg": "Try to get history files for current user",
            "results": []
        },
        "USER_BASH_HISTORY": {
            "cmd": 'cat ~/.bash_history ',
            "msg": "Try to get history files for current user",
            "results": []
        },
        "NANO_HISTORY": {
            "cmd": "cat ~/.nano_history",
            "msg": "Try to get the contents of nano history file for current "
                   "user",
            "results": []
        },
        "ATFTP_HISTORY": {
            "cmd": "cat ~/.atftp_history",
            "msg": "Try to get the contents of atftp history file for current "
                   "user",
            "results": []
        },
        "MYSQL_HISTORY": {
            "cmd": "cat ~/.mysql_history",
            "msg": "Try to get the contents of mysql history file for current "
                   "user",
            "results": []
        },
        "PHP_HISTORY": {
            "cmd": "cat ~/.php_history",
            "msg": "Try to get the contents of php history file for current "
                   "user",
            "results": []
        },
        "PYTHON_HISTORY": {
            "cmd": "cat ~/.python_history",
            "msg": "Try to get the contents of python history file for "
                   "current user",
            "results": []
        },
        "REDIS_HISTORY": {
            "cmd": "cat ~/.rediscli_history",
            "msg": "Try to get the contents of redis cli history file for "
                   "current user",
            "results": []
        },
        "TDSQL_HISTORY": {
            "cmd": "cat ~/.tdsql_history",
            "msg": "Try to get the contents of tdsql history file for current "
                   "user",
            "results": []
        }
    }

    command_list = list(historyfiles.keys())

    for item in command_list:
        cmd = historyfiles[item]['cmd']
        command_result = container.exec_run(
            cmd=cmd.split(' '),
            stderr=False,
            stdin=False)

        # If command is not executed successfully, dont show in output
        if command_result.exit_code != 0:
            del historyfiles[item]
            continue

        output = command_result.output.decode('utf-8')
        result = output.split('\n')
        historyfiles[item]['results'] = result
        del historyfiles[item]['cmd']

    result = historyfiles

    res[container.short_id] = {
        _plugin_name_: result
    }

    _log.info('Completed execution of {} Plugin.'.format(_plugin_name_))
    output_queue.put(res)

    if audit:
        _audit(container, historyfiles, audit_queue)


def _audit(container, scan_report, audit_queue):
    """Perform Scan audit.

    :param scan_report: dict
    :param audit_queue: Multiprocessing queue to perform Audit.
    """
    weak_configurations = {
        "ROOT_SH_HISTORY": {
            "warn_msg": "Root History present: ",
            "safe_conf": []
        },
        "ROOT_BASH_HISTORY": {
            "warn_msg": "Root History present: ",
            "safe_conf": []
        },
        "USER_SH_HISTORY": {
            "warn_msg": "USER BASH History present: ",
            "safe_conf": []
        },
        "USER_BASH_HISTORY": {
            "warn_msg": "USER BASH History present: ",
            "safe_conf": []
        },
        "NANO_HISTORY": {
            "warn_msg": "NANO History present: ",
            "safe_conf": []
        },
        "ATFTP_HISTORY": {
            "warn_msg": "ATFTP History present: ",
            "safe_conf": []
        },
        "MYSQL_HISTORY": {
            "warn_msg": "MySQL History present: ",
            "safe_conf": []
        },
        "PHP_HISTORY": {
            "warn_msg": "PHP History present: ",
            "safe_conf": []
        },
        "PYTHON_HISTORY": {
            "warn_msg": "Python History present: ",
            "safe_conf": []
        },
        "REDIS_HISTORY": {
            "warn_msg": "REDIS History present: ",
            "safe_conf": []
        },
        "TDSQL_HISTORY": {
            "warn_msg": "TDSQL History present: ",
            "safe_conf": []
        }
    }

    container_id = container.short_id
    audit_report = {}
    audit_report[container_id] = []
    columns = [_plugin_name_, 'WARN']

    for security_option in scan_report.keys():
        if security_option in weak_configurations.keys():
            actual_results = scan_report[security_option]['results']
            weak_results = None
            safe_results = None

            if 'weak_conf' in weak_configurations[security_option].keys():
                weak_results = weak_configurations[security_option]['weak_conf']

            if 'safe_conf' in weak_configurations[security_option].keys():
                safe_results = weak_configurations[security_option]['safe_conf']

            warn_message = weak_configurations[security_option]['warn_msg']

            # If safe_conf, check on safe_conf else check on weak_conf
            if safe_results is not None:
                if actual_results != safe_results:
                    _r = [res for res in actual_results if res]
                    if _r:
                        r = columns + [warn_message + ";".join(_r)]
                        audit_report[container_id].append(", ".join(r))
            else:
                if utils.list_intersection(actual_results, weak_results) != []:
                    # Weak configuration detected
                    r = columns + [warn_message]
                    audit_report[container_id].append(
                        ", ".join(r)
                    )

    audit_queue.put(audit_report)



"""Setup script."""
import DockerENT
import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

_version = DockerENT.__version__
_requires = open('requirements.txt').read().splitlines()

setuptools.setup(
    name="DockerENT",
    version=_version,
    description="A tool to analyse issues with running docker container(s)",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/r0hi7/DockerENT",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
    install_requires=_requires,
    entry_points={
        'console_scripts': {
            'DockerENT = DockerENT.__main__:start'
        }
    },

    keywords='docker runtime scanning framework'
)



"""Output worker, defines how to process output from each plugin."""
import importlib
import logging

_log = logging.getLogger(__name__)


def output_handler(queue, target, filename='out.json'):
    """Worker function passes the queue to target.

    :param queue: This queue holds the output from each plugin executed.
    :type queue: multiprocessing.managers.AutoProxy[Queue]

    :param target: Specifies which output plugin to write data to.
    :type target: DockerENT.output_plugins

    :return: None
    """
    output_plugin = importlib.import_module('DockerENT.output_plugins.'+target)
    output_plugin.write(queue, filename)



"""Docker user info scan plugin."""
from DockerENT.utils import utils
import logging

_log = logging.getLogger(__name__)

_plugin_name_ = 'user-info'


def scan(container, output_queue, audit=False, audit_queue=None):
    """Docker user-info scan plugin.

    :param container: container instance.
    :type container: docker.models.containers.Container

    :param output_queue: Output holder for this plugin.
    :type output_queue: multiprocessing.managers.AutoProxy[Queue]

    :return: This plugin returns the object in this form.
    {
        _plugin_name_: {
            'test_performed': {
                'results':  []
            }
        }
    }
    """
    res = {}

    _log.info('Staring {} Plugin ...'.format(_plugin_name_))
    userinfo = {
        "WHOAMI": {
            "cmd": "whoami",
            "msg": "Current User",
            "results": []
        },
        "ID": {
            "cmd": "id",
            "msg": "Current User ID",
            "results": []
        },
        "ALLUSERS": {
            "cmd": "cat /etc/passwd",
            "msg": "All users",
            "results": []
        },
        "ENV": {
            "cmd": "env",
            "msg": "Environment",
            "results": []
        },
        "SUDOERS": {
            "cmd": "cat /etc/sudoers",
            "msg": "Sudoers (privileged)",
            "results": []
        }
    }

    command_list = list(userinfo.keys())

    for item in command_list:
        cmd = "sh -c " + "\"" + userinfo[item]['cmd'] + "\""
        command_result = container.exec_run(
            cmd=cmd,
            stderr=True,
            stdin=False)

        # If command is not executed successfully, dont show in output
        if command_result.exit_code != 0:
            del userinfo[item]
            continue

        output = command_result.output.decode('utf-8')
        result = output.split('\n')
        userinfo[item]['results'] = result
        del userinfo[item]['cmd']

    result = userinfo

    res[container.short_id] = {
        _plugin_name_: result
    }

    _log.info('Completed execution of {} Plugin.'.format(_plugin_name_))
    output_queue.put(res)

    if audit:
        _audit(container, userinfo, audit_queue)


def _audit(container, scan_report, audit_queue):
    """Perform Scan audit.

    :param scan_report: dict
    :param audit_queue: Multiprocessing queue to perform Audit.
    """
    container_id = container.short_id
    audit_report = {}
    audit_report[container_id] = []

    columns = [_plugin_name_, 'WARN']

    weak_configurations = {
        "WHOAMI": {
            'weak_conf': ['root', 'admin', 'superuser'],
            'warn_msg': 'Current USER is : '
        },
        "ALLUSERS": {
            'safe_conf': [],
            'warn_msg': 'User with password in passwd file: '
        },
        "SUDOERS": {
            "safe_conf": [],
            "warn_message": 'Sudoers found : '
        }
    }

    for security_option in scan_report.keys():
        if security_option in weak_configurations.keys():
            actual_results = scan_report[security_option]['results']

            weak_results = None
            safe_results = None

            if 'weak_conf' in weak_configurations[security_option].keys():
                weak_results = weak_configurations[security_option]['weak_conf']

            if 'safe_conf' in weak_configurations[security_option].keys():
                safe_results = weak_configurations[security_option]['safe_conf']

            warn_message = weak_configurations[security_option]['warn_msg']

            # If safe_conf, check on safe_conf else check on weak_conf
            if safe_results is not None:
                if actual_results != safe_results:
                    if security_option == 'ALLUSERS':
                        plain_text_creds = []

                        for ar in actual_results:
                            result_list = ar.split(':')
                            if len(result_list) > 1 and (
                                    result_list[1] != '!' and
                                    result_list[1] != '*' and
                                    result_list[1] != 'x'):
                                plain_text_creds.append(ar)

                        if plain_text_creds:
                            r = columns + [warn_message + ";".join(
                                plain_text_creds
                            )]
                            audit_report[container_id].append(", ".join(r))
            else:
                issues = utils.list_intersection(
                    actual_results, weak_results)
                if issues != []:
                    # Weak configuration detected
                    r = columns + [warn_message + ";".join(issues)]
                    audit_report[container_id].append(
                        ", ".join(r)
                    )

    audit_queue.put(audit_report)



"""Helper functions goes here."""

import copy


def parse_output_queue_2_dict(output_q):
    """Utility function to parse the output queue to python dict."""
    report = {}
    while not output_q.empty():
        result = output_q.get()
        for key in result.keys():
            if key in report.keys():
                report[key].update(result[key])
            else:
                report[key] = {}
                report[key].update(result[key])

    return report


def parse_audit_queue_2_dict(audit_q):
    """Utility function to parse the audit queue to python dict."""
    audit_report = {}
    while not audit_q.empty():
        result = audit_q.get()
        for key in result.keys():
            if not key in audit_report.keys():
                audit_report[key] = []

            audit_report[key].extend(result[key])
    return audit_report


def get_value_from_str_dotted_key(d, dotted_key):
    """Get value from python dict via dotted notation.

    :param d: dictionary
    :param dotted_key:

    :return: str
    """
    keys = dotted_key.split('.')
    temp = copy.deepcopy(d)
    try:
        for key in keys:
            temp = temp[key]
        return temp
    except KeyError:
        return None


def list_intersection(lst1, lst2):
    """List intersection."""
    lst3 = [value for value in lst1 if value in lst2]
    return lst3


def docker_network_response_parser(port_mapping_list):
    """Parse list to source - dest port.

    Input:
        [
            {
                '8080/tcp': [{'HostIp': '0.0.0.0', 'HostPort': '80'}],
                '90/tcp': [{'HostIp': '0.0.0.0', 'HostPort': '9090'}]
            }
        ]

    Output:
        ["HostIp(0.0.0.0):HostPort(80):GuestPort(8080)"]
    """
    result = []
    port_map = port_mapping_list[0]
    for pm in port_map.keys():
        result.append(
            "HostIp("+port_map[pm][0]['HostIp'] +
            "):HostPort("+port_map[pm][0]['HostPort'] +
            "):GuestPort("+pm+")"
        )
    return result



"""Start scanner workers.

This is scanner worker, this will trigger the plugin based scan on docker and
on docker-network which are specified by the controller.
"""
from DockerENT import docker_nw_plugins
from DockerENT import docker_plugins

import docker
import DockerENT
import importlib
import logging
import pkgutil

_log = logging.getLogger(__name__)

docker_client = docker.from_env()


def executor(target,
             plugin,
             output_queue,
             is_docker=False,
             audit=False,
             audit_queue=None):
    """Execute a plugin on the target.

    Target is either a docker.models.containers.Container or
    docker.models.networks.Network.
    The plugin will be either from DockerENT.docker_plugins or
    DockerENT.docker_nw_plugins depending upon the target type.

    :param target: Target to run plugin on.
    :type target: docker.models.containers.Container or
        docker.models.networks.Network

    :param plugin: Plugin.
    :type plugin: DockerENT.docker_plugins or
        DockerENT.docker_nw_plugins

    :param output_queue: Queue to hold output from each plugin.
    :type output_queue: multiprocessing.managers.AutoProxy[Queue]

    :param is_docker: specify is the target is docker or docker-nw instance.
    :type is_docker: bool

    :return: None
    """
    # load the plugin module
    if is_docker:
        package = docker_plugins.__package__
        target = docker_client.containers.get(target)
    else:
        package = docker_nw_plugins.__package__
        target = docker_client.networks.get(target)

    module = importlib.import_module(package + '.' + plugin)
    module.scan(target, output_queue, audit, audit_queue)


def docker_scan_worker(containers, plugins, process_pool, output_queue,
                       audit, audit_queue):
    """Docker scan worker.

    :param containers: Containers to scan.
    :type containers: List[docker.models.containers.Container]

    :param plugins: List of plugins..
    :type plugins: List[DockerENT.docker_plugins]

    :param process_pool: The multiprocessing pool object.
    :type process_pool: multiprocessing.pool.Pool

    :param output_queue: Queue to hold data from each plugin after execution.
    :type output_queue: multiprocessing.managers.AutoProxy[Queue]

    :return: None
    """
    _containers = []
    if containers == 'all' or containers[0] == 'all':
        _containers = docker_client.containers.list()
    else:
        _containers.append(docker_client.containers.get(containers))
    print(_containers)
    _plugins = []
    if plugins is None or plugins == 'all' or plugins[0] == 'all':
        for importer, modname, ispkg in pkgutil.iter_modules(
                DockerENT.docker_plugins.__path__):
            _plugins.append(modname)
    else:
        _plugins.append(plugins)

    plugins = _plugins
    containers = _containers

    _log.info('{} docker containers loaded ...'.format(len(containers)))
    _log.info('{} docker plugin(s) loaded ...'.format(len(plugins)))

    executor_args = []
    for container in containers:
        for plugin in plugins:
            executor_args.append(
                (container.id, plugin, output_queue, True, audit, audit_queue))

    _log.debug(executor_args)

    process_pool.starmap(executor, executor_args)


def docker_nw_scan_worker(nws, plugins, process_pool, output_queue):
    """Docker Network scan worker.

    :param nws: Docker networks to scan.
    :type nws: List[docker.models.networks.Network]

    :param plugins: List of plugins..
    :type plugins: List[DockerENT.docker_nw_plugins]

    :param process_pool: The multiprocessing pool object.
    :type process_pool: multiprocessing.pool.Pool

    :param output_queue: Queue to hold data from each plugin after execution.
    :type output_queue: multiprocessing.managers.AutoProxy[Queue]

    :return: None
    """
    _nws = []
    if nws == 'all':
        _nws = docker_client.networks.list()
    else:
        _nws.append(docker_client.networks.get(nws))

    _plugins = []
    if plugins is None or plugins == 'all':
        for importer, modname, ispkg in pkgutil.iter_modules(
                DockerENT.docker_nw_plugins.__path__):
            _plugins.append(modname)
    else:
        _plugins.append(plugins)

    plugins = _plugins
    nws = _nws

    _log.info('{} docker nws loaded ...'.format(len(nws)))
    _log.info('{} docker-network plugin(s) loaded ...'.format(
        len(plugins)))

    executor_args = []
    for nw in nws:
        for plugin in plugins:
            executor_args.append((nw.id, plugin, output_queue, False,))

    _log.debug(executor_args)

    process_pool.starmap_async(executor, executor_args)



"""Docker cron jobs scan plugin."""
import logging

_log = logging.getLogger(__name__)

_plugin_name_ = 'cron-info'


def scan(container, output_queue, audit=False, audit_queue=None):
    """Docker cron-jobs scan plugin.

    :param container: container instance.
    :type container: docker.models.containers.Container

    :param output_queue: Output holder for this plugin.
    :type output_queue: multiprocessing.managers.AutoProxy[Queue]

    :return: This plugin returns the object in this form.
    {
        _plugin_name_: {
            'test_performed': {
                'results':  []
            }
        }
    }
    """
    res = {}

    _log.info('Staring {} Plugin ...'.format(_plugin_name_))

    croninfo = {
        "user-cron-jobs": {
            "cmd": "crontab -l 2>/dev/null",
            "msg": "Users cron jobs",
            "results": []
        }
    }

    for item in croninfo:
        cmd = croninfo[item]['cmd']
        output = container.exec_run(cmd).output.decode('utf-8')
        result = output.split('\n')
        croninfo[item]['results'] = result
        del croninfo[item]['cmd']

    result = croninfo

    res[container.short_id] = {
        _plugin_name_: result
    }

    _log.info('Completed execution of {} Plugin.'.format(_plugin_name_))
    output_queue.put(res)
